/*******************************************************************************

        @file DisplayWriter.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version; March 2004      
                        Changed from sprintf() to mango.format; March 2005

        @author         Kris
                        Anders F Bjorklund (Darwin patches)
                        

*******************************************************************************/

module mango.io.DisplayWriter;

public  import  mango.io.Writer;

private import  mango.format.Int,
                mango.format.Long,
                mango.format.Double,
                mango.format.Number;


/*******************************************************************************

        Format output suitable for presentation. There is no corresponding
        reader, since there are no explicit delimiters emitted to isolate
        each formatted item. One can use TextReader to read input from the 
        console (or from some other input buffer).

*******************************************************************************/

class DisplayWriter : Writer
{
        alias Number.Radix      Radix;

        // temporary buffer space for encoding
        private char[64]        tmp;

        private char            fill;

        // number of floating-point digits to write 
        private uint            digits = 2;

        // how to print unsigned ints and unsigned longs
        private int             width = tmp.length;
        private Radix           radix = Radix.Decimal;

        /***********************************************************************
        
                Construct a DisplayWriter upon the specified IBuffer

        ***********************************************************************/

        this (IBuffer buffer)
        {
                super (buffer);
                
                // disable array length output (for strings and so on)
                enableArrayPrefix (false);

                encode.int1 = &int1;
                encode.int8 = &int8;
                encode.int8u = &int8u;
                encode.int16 = &int16;
                encode.int16u = &int16u;
                encode.int32 = &int32;
                encode.int32u = &int32u;
                encode.int64 = &int64;
                encode.int64u = &int64u;

                encode.float32 = &float32;
                encode.float64 = &float64;
                encode.float80 = &float80;
        }
     
        /***********************************************************************
        
                Construct a DisplayWriter upon the specified IConduit

        ***********************************************************************/

        this (IConduit conduit)
        {
                this (conduit.createBuffer);
        }

        /***********************************************************************
        
                Set the number of floating point digits

        ***********************************************************************/

        void setPrecision (uint digits)
        {
                this.digits = digits;
        }

        /***********************************************************************
        
        ***********************************************************************/

        void setRadix (Radix radix)
        {
                this.radix = radix;
        }

        /***********************************************************************
        
        ***********************************************************************/

        void setFill (uint width, char fill = ' ')
        {
                assert (width <= tmp.length);
                this.width = width;
                this.fill = fill;
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void formatInt (uint i, bool sign)
        {     
                if (sign)
                    put (Int.format (tmp, i));
                else
                   put (Int.format (tmp[0..width], i, radix, fill));
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void formatLong (ulong i, bool sign)
        {     
                if (sign)
                    put (Long.format (tmp, i));
                else
                   put (Long.format (tmp[0..width], i, radix, fill));
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void int8f (ubyte* src, uint count, bool sign)
        {     
                while (count)
                      {
                      // sign extend!
                      int i = *src;
                      formatInt (i, sign);

                      ++src;
                      --count;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void int16f (ushort* src, uint count, bool sign)
        {     
                while (count)
                      {
                      // sign extend!
                      int i = *src;
                      formatInt (i, sign);

                      ++src;
                      count -= 2;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void int32f (uint* src, uint count, bool sign)
        {     
                while (count)
                      {
                      uint i = *src;
                      formatInt (i, sign);

                      ++src;
                      count -= 4;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void int64f (ulong* src, uint count, bool sign)
        {     
                while (count)
                      {
                      ulong i = *src;
                      formatLong (i, sign);

                      ++src;
                      count -= 8;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        private void fformat (void* src, uint count, uint width)
        {     
                double x;

                while (count)
                      {
                      if (width == float.sizeof)
                          x = *cast(float*) src;
                      else
                      if (width == double.sizeof)
                          x = *cast(double*) src;
                      else
                          x = *cast(real*) src;
                      
                      put (Double.format (tmp, x, digits));    
                        
                      count -= width;
                      src += width;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int1 (void* src, uint count)
        {     
                while (count)
                      {
                      if (*cast(ubyte *) src)
                          encode.char8 (cast(char*) "true", 4);
                      else
                         encode.char8 (cast(char*) "false", 5);

                      ++src;
                      --count;
                      }
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int8 (void* src, uint count)
        {     
                int8f (cast(ubyte*) src, count, true); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int8u (void* src, uint count)
        {     
                int8f (cast(ubyte*) src, count, false); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int16 (void* src, uint count)
        {     
                int16f (cast(ushort*) src, count, true); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int16u (void* src, uint count)
        {     
                int16f (cast(ushort*) src, count, false); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int32 (void* src, uint count)
        {     
                int32f (cast(uint*) src, count, true); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int32u (void* src, uint count)
        {     
                int32f (cast(uint*) src, count, false); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int64 (void* src, uint count)
        {     
                int64f (cast(ulong*) src, count, true); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void int64u (void* src, uint count)
        {     
                int64f (cast(ulong*) src, count, false); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void float32 (void* src, uint count)
        {     
                fformat (src, count, float.sizeof); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void float64 (void* src, uint count)
        {     
                fformat (src, count, double.sizeof); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final void float80 (void* src, uint count)
        {     
                fformat (src, count, real.sizeof); 
        }
}



/*******************************************************************************

        Subclass to support automatic flushing. This can be used for 
        Stdout, Stderr, and other related conduits.

*******************************************************************************/

class FlushWriter : DisplayWriter
{
        alias DisplayWriter.put put;

        /***********************************************************************
        
                Construct a FlushWriter upon the specified IBuffer

        ***********************************************************************/

        this (IBuffer buffer)
        {
                super (buffer);
        }

        /***********************************************************************
        
                Construct a FlushWriter upon the specified IConduit

        ***********************************************************************/

        this (IConduit conduit)
        {
                this (conduit.createBuffer);
        }

        /**********************************************************************
        
                Intercept the IWritable method to catch newlines, and 
                flush the buffer whenever one is emitted

        ***********************************************************************/

        override IWriter put (IWritable x)
        {      
               // have superclass handle the IWritable
                super.put (x);

                // flush output when we see a newline
                if (cast(INewlineWriter) x)
                    flush ();

                return this;
        }
}
     
