/*******************************************************************************

        @file FileConduit.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris 
                        John Reimer
                        Anders F Bjorklund (Darwin patches)
                        Chris Sauls (Win95 file support)

*******************************************************************************/

module mango.io.FileConduit;

public  import  mango.base.System;

public  import  mango.io.Buffer,
                mango.io.FilePath,
                mango.io.FileProxy,
                mango.io.FileStyle;

private import  mango.io.Conduit,
                mango.io.Exception,
                mango.io.ConduitStyle;

/*******************************************************************************

        Bring in native Unix functions. Note that the functions read()
        and close() will actually hide the inherited Conduit.read() and 
        Resource.close() if the import is placed within the body of the
        FileConduit class. Apparently this behaviour is 'by design', so
        fair warning to those who might expect otherwise.

*******************************************************************************/

version (Posix)
        {
        version (linux) 
                {
                private import std.c.linux.linux;
                alias std.c.linux.linux posix;
                }

        version (darwin) 
                {
                private import std.c.darwin.darwin;
                alias std.c.darwin.darwin posix;
                }
                
        private extern (C) int ftruncate (int, int);
        }

/*******************************************************************************

        Bring in native Windows functions

*******************************************************************************/

version (Win32)
        {
        private import std.c.windows.windows;  
             
        private extern (Windows) 
        {
        HANDLE GetStdHandle (DWORD);
        BOOL   SetEndOfFile (HANDLE);

        BOOL   UnmapViewOfFile    (LPCVOID);
        BOOL   FlushViewOfFile    (LPCVOID, DWORD);
        LPVOID MapViewOfFile      (HANDLE, DWORD, DWORD, DWORD, DWORD);
        HANDLE CreateFileMappingA (HANDLE, LPSECURITY_ATTRIBUTES, DWORD, DWORD, DWORD, LPCTSTR);
        }
        }


/*******************************************************************************

        Implements a means of reading and writing a generic file. Conduits
        are the primary means of accessing external data, and are usually 
        routed through a Buffer. File conduit extends the generic conduit
        by providing file-specific methods to set the file size, seek to a
        specific file position, and so on. Also provided is a factory for
        create a memory-mapped Buffer upon a file.     
        
        Serial input and output is straightforward. In this example we
        copy a file directly to the console:

        @code
        // open a file for reading
        FileConduit fc = new FileConduit ("test.txt");

        // stream directly to console
        Stdio.stdout.copy (fc);
        @endcode

        And here we copy one file to another:

        @code
        // open a file for reading
        FileConduit from = new FileConduit ("test.txt");

        // open another for writing
        FileConduit to = new FileConduit ("copy.txt", FileStyle.WriteTruncate);

        // copy file
        to.copy (from);
        @endcode
        
        FileConduit can just as easily handle random IO. Here we see how
        a Reader and Writer are used to perform simple input and output:

        @code
        // open a file for reading
        FileConduit fc = new FileConduit ("random.bin", FileStyle.ReadWriteCreate);

        // construct (binary) reader & writer upon this conduit
        Reader r = new Reader(fc);
        Writer w = new Writer(fc);

        int x=10, y=20;

        // write some data
        w.put(x).put(y);

        // flush output since IO is buffered
        w.flush();

        // rewind to file start
        fc.seek (0);

        // read data back again, but swap destinations
        r.get(y).get(x);

        assert (y==10);
        assert (x==20);

        fc.close();
        @endcode

        See FilePath, FileProxy, and FileSystem for additional functionality
        related to file manipulation. Note that classes CompositeWriter and
        CompositeReader may be more appropriate for the writing & reading of
        structured data. Doxygen has a hard time with D version() statements, 
        so part of this class is documented within FileConduit::VersionWin32 
        instead.

        Compile with -version=Win32SansUnicode to enable Win95 & Win32s file 
        support.
        
*******************************************************************************/

class FileConduit : Conduit, ISeekable
{
        // the file we're working with 
        private FilePath path;

        // expose conduit copy() method also (alias is truly non-intuitive ...)
        alias Conduit.copy copy;

        /***********************************************************************
        
                Create a FileConduit with the provided path and style.

        ***********************************************************************/

        this (char[] name, FileStyle style = FileStyle.ReadExisting)
        {
                this (new FilePath(name), style);
        }

        /***********************************************************************
        
                Create a FileConduit from the provided proxy and style.

        ***********************************************************************/

        this (FileProxy proxy, FileStyle style = FileStyle.ReadExisting)
        {
                this (proxy.getPath(), style);
        }

        /***********************************************************************
        
                Create a FileConduit with the provided path and style.

        ***********************************************************************/

        this (FilePath path, FileStyle style = FileStyle.ReadExisting)
        {
                // say we're seekable
                super (style, true);
                
                // remember who we are
                this.path = path;

                // open the file
                _open (style);

                // bump lock count
                acquire();
        }    

        /***********************************************************************

                Create a FileConduit on the provided FileDevice. This is 
                strictly for adapting existing devices such as Stdout and
                friends.
        
        ***********************************************************************/

        package this (FileDevice device)
        {
                // say we're not seekable
                super (device, false);
                
                // devices don't have a path (bug?)
                this.path = null;

                // open the file
                _reopen (device);

                // bump lock count
                acquire();
        }    

        /***********************************************************************
        
                Callback to close the file. This is invoked from the Resource
                base-class when the resource is being closed.

        ***********************************************************************/

        protected override void closure ()
        {       
                super.closure ();
                _close ();
        }    
                   
        /***********************************************************************
        
                Return the FilePath used by this file.

        ***********************************************************************/

        FilePath getPath ()
        {
                return path;
        }               

        /***********************************************************************
        
                Move the file position to the given offset (from the file 
                start) and return the adjusted postion.

        ***********************************************************************/

        long seek (long offset)
        {
                return seek (offset, SeekAnchor.Begin);
        }               

        /***********************************************************************
                
                Return the current file position.
                
        ***********************************************************************/

        long getPosition ()
        {
                return seek (0, SeekAnchor.Current);
        }               


        /***********************************************************************

                Create a Buffer of the default FileConduit size, and
                associate it with this conduit

        ***********************************************************************/

        override IBuffer createBuffer()
        {
                const int DefaultFileBufferSize = 1024 * 16; 

                IBuffer buffer = new Buffer (DefaultFileBufferSize);

                buffer.setConduit (this);
                return buffer;
        } 

        /***********************************************************************
        
                Return the total length of this file.

        ***********************************************************************/

        long length ()
        {
                long    pos,    
                        ret;
                        
                pos = getPosition ();
                ret = seek (0, SeekAnchor.End);
                seek (pos);
                return ret;
        }               

        /***********************************************************************

                Transfer the content of another file to this one. Returns a
                reference to this class on success, or throws an IOException 
                upon failure.
        
        ***********************************************************************/

        FileConduit copy (FilePath source)
        {
                auto FileConduit fc = new FileConduit (source);
                super.copy (fc);
                return this;
        }               

        /***********************************************************************

                Windows-specific code
        
        ***********************************************************************/

        version(Win32)
        {
                private HANDLE          handle;
                private bool            console, 
                                        appending;

                /***************************************************************

                        Throw an IOException noting the last error

                ***************************************************************/

                private void exception ()
                {
                        char[] name = (path) ? path.toString : "console";

                        throw new IOException (name ~ ": " ~ System.error);
                }

                /***************************************************************

                        Gain access to the standard IO handles (console etc).

                ***************************************************************/

                private void _reopen (FileDevice device)
                {
                        static const DWORD[] id = [
                                                  cast(DWORD) -10, 
                                                  cast(DWORD) -11, 
                                                  cast(DWORD) -12
                                                  ];
                        static const char[][] f = [
                                                  "CONIN$\0", 
                                                  "CONOUT$\0", 
                                                  "CONOUT$\0"
                                                  ];

                        handle = GetStdHandle (id[device.id]);
                        if (! handle)
                              handle = CreateFileA (f[device.id], 
                                       GENERIC_READ | GENERIC_WRITE,  
                                       FILE_SHARE_READ | FILE_SHARE_WRITE, 
                                       null, OPEN_EXISTING, 0, null);
                        if (! handle)
                              exception();
                        console = true;
                }

                /***************************************************************

                        Open a file with the provided style.

                ***************************************************************/

                private void _open (FileStyle style)
                {
                        DWORD   attr,
                                share,
                                access,
                                create;

                        alias DWORD[] Flags;

                        static const Flags Access =  
                                        [
                                        0,                      // invalid
                                        GENERIC_READ,
                                        GENERIC_WRITE,
                                        GENERIC_READ | GENERIC_WRITE,
                                        ];
                                                
                        static const Flags Create =  
                                        [
                                        OPEN_EXISTING,          // must exist
                                        CREATE_ALWAYS,          // create always
                                        TRUNCATE_EXISTING,      // must exist
                                        CREATE_ALWAYS,          // (for appending)
                                        ];
                                                
                        static const Flags Share =   
                                        [
                                        FILE_SHARE_READ,
                                        FILE_SHARE_WRITE,
                                        FILE_SHARE_READ | FILE_SHARE_WRITE,
                                        ];
                                                
                        static const Flags Attr =   
                                        [
                                        0,
                                        FILE_FLAG_RANDOM_ACCESS,
                                        FILE_FLAG_SEQUENTIAL_SCAN,
                                        0,
                                        FILE_FLAG_WRITE_THROUGH,
                                        ];

                        attr = Attr[style.cache];
                        share = Share[style.share];
                        create = Create[style.open];
                        access = Access[style.access];

                        version (Win32SansUnicode)
                                 handle = CreateFileA (path.toUtf8, access, share, 
                                                       null, create, 
                                                       attr | FILE_ATTRIBUTE_NORMAL,
                                                       cast(HANDLE) null);
                             else
                                handle = CreateFileW (path.toUtf16, access, share, 
                                                      null, create, 
                                                      attr | FILE_ATTRIBUTE_NORMAL,
                                                      cast(HANDLE) null);

                        if (handle == INVALID_HANDLE_VALUE)
                            exception ();

                        // move to end of file?
                        if (style.open == style.Open.Append)
                            appending = true;
                }
                
                /***************************************************************

                        Close the underlying file

                ***************************************************************/

                private void _close ()
                {
                        if (! CloseHandle (handle))
                              exception ();
                }

                /***************************************************************

                        Read a chunk of bytes from the file into the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                protected override int reader (void[] dst)
                {
                        DWORD read;
                        void *p = dst;

                        if (! ReadFile (handle, p, dst.length, &read, null))
                              exception ();

                        if (read == 0 && dst.length > 0)
                            return Eof;
                        return read;
                }

                /***************************************************************

                        Write a chunk of bytes to the file from the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                protected override int writer (void[] src)
                {
                        const int Max = 0x7fff;
                        DWORD written = src.length;

                        // Win32 console has problems with content over
                        // 32KB in length, so handle it specifically
                        if (console && written > Max)
                           {
                           DWORD len = written;
                           do {
                              DWORD i = len;
                              if (i > Max)
                                  i = Max;
                              if (! WriteFile (handle, src, i, &i, null))
                                    exception ();
                              src = src [i .. length];
                              len -= i;
                              } while (len);
                           }
                        else
                           {
                           // try to emulate the Unix O_APPEND mode
                           if (appending)
                               SetFilePointer (handle, 0, null, SeekAnchor.End);

                           if (! WriteFile (handle, src, src.length, &written, null))
                                 exception ();
                           }

                        return written;
                }
                
                /***************************************************************

                        Set the file size to be that of the current seek 
                        position. The file must be writable for this to
                        succeed.

                ***************************************************************/

                void truncate ()
                {
                        // must have Generic_Write access
                        if (! SetEndOfFile (handle))
                              exception ();                            
                }               

                /***************************************************************

                        Set the file seek position to the specified offset
                        from the given anchor. 

                ***************************************************************/

                long seek (long offset, SeekAnchor anchor)
                {
                        LONG high = (offset >> 32);
                        long result = SetFilePointer (handle, cast(LONG) offset, 
                                                      &high, anchor);

                        if (result == -1 && 
                            GetLastError() != ERROR_SUCCESS)
                            exception ();

                        return result + (cast(long) high << 32);
                }               
        }


        /***********************************************************************

                 Unix-specific code. Note that some methods are 32bit only
        
        ***********************************************************************/

        version (Posix)
        {
                private int handle = -1;

                /***************************************************************

                        Throw an IOException noting the last error

                ***************************************************************/

                private void exception ()
                {
                        throw new IOException (path.toString() ~ ": " ~
                                               System.error);
                }

                /***************************************************************

                        Gain access to the standard IO handles (console etc).

                ***************************************************************/

                private void _reopen (FileDevice device)
                {
                        handle = device.id;
                }

                /***************************************************************

                        Open a file with the provided style.

                ***************************************************************/

                private void _open (FileStyle style)
                {
                        int     share, 
                                access;

                        alias int[] Flags;

                        static const Flags Access =  
                                        [
                                        0,              // invalid
                                        O_RDONLY,
                                        O_WRONLY,
                                        O_RDWR,
                                        ];
                                                
                        static const Flags Create =  
                                        [
                                        0,              // open existing
                                        O_CREAT,        // create always
                                        O_TRUNC,        // must exist
                                        O_APPEND | O_CREAT, 
                                        ];

                        // this is not the same as Windows sharing,
                        // but it's perhaps a reasonable approximation
                        static const Flags Share =   
                                        [
                                        0640,           // read access
                                        0620,           // write access
                                        0660,           // read & write
                                        ];
                                                
                        share = Share[style.share];
                        access = Access[style.access] | Create[style.open];

                        handle = posix.open (path.toUtf8, access, share);
                        if (handle == -1)
                            exception ();
                }

                /***************************************************************

                        Close the underlying file

                ***************************************************************/

                private void _close ()
                {
                        if (posix.close (handle) == -1)
                            exception ();
                }

                /***************************************************************

                        Read a chunk of bytes from the file into the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                protected override int reader (void[] dst)
                {
                        int read = posix.read (handle, dst, dst.length);
                        if (read == -1)
                            exception ();
                        else
                           if (read == 0 && dst.length > 0)
                               return Eof;
                        return read;
                }

                /***************************************************************

                        Write a chunk of bytes to the file from the provided
                        array (typically that belonging to an IBuffer)

                ***************************************************************/

                protected override int writer (void[] src)
                {
                        int written = posix.write (handle, src, src.length);
                        if (written == -1)
                            exception ();
                        return written;
                }
                
                /***************************************************************

                        32bit only ...

                        Set the file size to be that of the current seek 
                        position. The file must be writable for this to
                        succeed.

                ***************************************************************/

                void truncate ()
                {
                        // set filesize to be current seek-position
                        if (ftruncate (handle, getPosition()) == -1)
                            exception ();
                }               

                /***************************************************************

                        32bit only ...

                        Set the file seek position to the specified offset
                        from the given anchor. 

                ***************************************************************/

                long seek (long offset, SeekAnchor anchor)
                {
                        int result = posix.lseek (handle, offset, anchor);
                        if (result == -1)
                            exception ();
                        return result;
                }               
        }
}


/*******************************************************************************

        Class to handle memory-mapped files

*******************************************************************************/

class MappedFile : MappedBuffer
{
        version (Windows)
                {
                private FileConduit     host;
                private void*           base;
                private HANDLE          mmFile;

                /***************************************************************

                        Construct a MappedBuffer upon the given FileConduit. 
                        One should set the file size using seek() & truncate() 
                        to setup the available working space.

                ***************************************************************/

                this (FileConduit host)
                {
                        this.host = host;

                        // can only do 32bit mapping on 32bit platform
                        long size = host.length;
                        ConduitStyle style = host.getStyle;

                        DWORD flags = PAGE_READWRITE;
                        if (style.access == ConduitStyle.Access.Read)
                            flags = PAGE_READONLY;

                        mmFile = CreateFileMappingA (host.handle, null, flags, 0, 0, null);
                        if (mmFile is null)
                            host.exception ();

                        flags = FILE_MAP_WRITE;
                        if (style.access == ConduitStyle.Access.Read)
                            flags = FILE_MAP_READ;

                        base = MapViewOfFile (mmFile, flags, 0, 0, 0);
                        if (base is null)
                            host.exception ();

                        void[] mem = base[0..size];
                        setValidContent (mem);
                }

                /***************************************************************

                        Ensure this is closed when GC'd

                ***************************************************************/

                ~this ()
                {
                        close();        
                }

                /***************************************************************

                        Close this mapped buffer

                ***************************************************************/

                void close ()
                {
                        if (base)
                            UnmapViewOfFile (base);

                        if (mmFile)
                            CloseHandle (mmFile);       

                        mmFile = null;
                        base = null;
                }

                /***************************************************************

                        Flush dirty content out to the drive. This
                        fails with error 33 if the file content is
                        virgin. Opening a file for ReadWriteExists
                        followed by a flush() will cause this.

                ***************************************************************/

                void flush ()
                {
                        // flush all dirty pages
                        if (! FlushViewOfFile (base, 0))
                              host.exception();
                }
                }


        version (Posix)
                {
                // not yet supported
                private this (){};
                }
}


/*******************************************************************************

        Class used to wrap an existing file-oriented handle, such as Stdout
        and its cohorts.

*******************************************************************************/

class FileDevice : ConduitStyle
{
        private uint id;

        this (uint id, Access access)
        in {
           assert (id < 3);
           }
        body
        {
                super (access);
                this.id = id;
        }
}


