/*******************************************************************************

        @file Reader.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.Reader;

public import mango.io.AbstractReader;

/*******************************************************************************

        Binary data Reader. Each reader operates upon an IBuffer, which is
        provided at construction time. Said buffer is intended to remain 
        consistent over the reader lifetime.

        Readers support both a C++ iostream type syntax, along with a get()
        syntax. Operations may be chained back-to-back.

        All readers support the full set of native data types, plus a
        selection of array types. The latter can be configured to produce
        either a copy (.dup) of the buffer content, or a slice. See class
        SimpleAllocator, BufferAllocator and SliceAllocator for more on 
        this topic.

        The code below illustrates basic operation upon a memory buffer:
        
        @code
        Buffer buf = new Buffer (256);

        // map same buffer into both reader and writer
        IReader r = new Reader(buf);
        IWriter w = new Writer(buf);

        int i = 10;
        long j = 20;
        double d = 3.14159;
        char[] c = "fred";

        // write data types out
        w << c << i << j << d;

        // read them back again
        r >> c >> i >> j >> d;

        // reset
        buf.clear();

        // same thing again, but using get() syntax instead
        w.put(c).put(i).put(j).put(d);
        r.get(c).get(i).get(j).get(d);

        @endcode

        Readers may also be used with any class implementing the IReadable
        interface. See PickleReader for an example of how this can be put
        to good use.
        
*******************************************************************************/

class Reader : AbstractReader
{       
        /***********************************************************************
        
                Construct a reader on the provided buffer.

        ***********************************************************************/

        this (IBuffer buffer)
        {
                super (buffer);

                decode.int1    = 
                decode.int8    = 
                decode.int8u   =
                decode.int16   = 
                decode.int16u  = 
                decode.int32   = 
                decode.int32u  =
                decode.int64   =
                decode.int64u  =
                decode.float32 =
                decode.float64 =
                decode.float80 =
                decode.char8    = 
                decode.char16   = 
                decode.char32   = &read;
        }

        /***********************************************************************
        
                Construct a reader on the buffer belonging to the given
                conduit.

        ***********************************************************************/

        this (IConduit conduit)
        {
                this (conduit.createBuffer); 
        }

        /***********************************************************************
        
        ***********************************************************************/

        final uint read (void *dst, uint count)
        {
                uint i = count;
                while (i)
                      {
                      // get as much as there is available in the buffer
                      uint available = buffer.readable();
                      
                      // cap bytes read
                      if (available > i)
                          available = i;

                      // copy them over
                      dst[0..available] = buffer.get (available);

                      // bump counters
                      dst += available;
                      i -= available;

                      // if we need more, prime the input by reading
                      if (i)
                          buffer.get(1, false);
                      }
                return count;
        }
}


