/*******************************************************************************

        @file TextReader.d
        
        Copyright (c) 2004 Kris Bell
        
        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.
        
        Permission is hereby granted to anyone to use this software for any 
        purpose, including commercial applications, and to alter it and/or 
        redistribute it freely, subject to the following restrictions:
        
        1. The origin of this software must not be misrepresented; you must 
           not claim that you wrote the original software. If you use this 
           software in a product, an acknowledgment within documentation of 
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must 
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, March 2004      
        @author         Kris


*******************************************************************************/

module mango.io.TextReader;

private import  mango.io.Token,
                mango.io.Tokenizer,
                mango.io.Exception,
                mango.io.AbstractReader;

/*******************************************************************************

        Grok readable input from a stream. All input is tokenized from the
        associated buffer, and converted as necessary into the destination 
        location. A SpaceTokenizer is used by default, but you can choose
        an alternative (such as a comma-delimiting tokenizer).

*******************************************************************************/

class TextReader : AbstractReader
{
        alias AbstractReader.get get;

        // an individual token
        private Token token;

        // internal tokenizer instance
        private ITokenizer tokenizer;

        /***********************************************************************
        
                Construct a TextReader on the provided buffer, using the
                specified Tokenizer instead of the default one.

        ***********************************************************************/

        this (IBuffer buffer, ITokenizer tokenizer = Tokenizers.line)
        {
                super (buffer);
                this.token = new Token;
                this.tokenizer = tokenizer;
     
                decode.int1 = &int1;
                decode.int8 = 
                decode.int8u = &int8;
                decode.int16 =
                decode.int16u = &int16;
                decode.int32 = 
                decode.int32u = &int32;
                decode.int64 = 
                decode.int64u = &int64;

                decode.float32 = &float32;
                decode.float64 = &float64;
                decode.float80 = &float80;
        }
        
        /***********************************************************************
                
                Construct a TextReader upon the buffer associated with the
                given conduit.

        ***********************************************************************/

        this (IConduit conduit, ITokenizer tokenizer = Tokenizers.line)
        {
                this (conduit.createBuffer, tokenizer);
        }

        /***********************************************************************
        
                Internal method to isolate the next token.

        ***********************************************************************/

        private final Token next ()
        {
                if (tokenizer.next (buffer, token) && token.getLength > 0)
                    return token;    
                throw new TokenException ("unexpected end of input");
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint int1 (void* src, uint count)
        {     
                for (uint i=count; i; --i)
                    {
                    * cast(bool*) src = (next.toString == "true");
                    ++src;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint int8 (void* src, uint count)
        {     
                for (uint i=count; i; i -= byte.sizeof)
                    {
                    * cast(byte*) src = next.toInt();
                    src += byte.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint int16 (void* src, uint count)
        {     
                for (uint i=count; i; i -= short.sizeof)
                    {
                    * cast(short*) src = next.toInt();
                    src += short.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint int32 (void* src, uint count)
        {     
                for (uint i=count; i; i -= int.sizeof)
                    {
                    * cast(int*) src = next.toInt();
                    src += int.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint int64 (void* src, uint count)
        {     
                for (uint i=count; i; i -= long.sizeof)
                    {
                    * cast(long*) src = next.toLong();
                    src += long.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint float32 (void* src, uint count)
        {     
                for (uint i=count; i; i -= float.sizeof)
                    {
                    * cast(float*) src = next.toReal();
                    src += float.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint float64 (void* src, uint count)
        {     
                for (uint i=count; i; i -= double.sizeof)
                    {
                    * cast(double*) src = next.toReal();
                    src += double.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        uint float80 (void* src, uint count)
        {     
                for (uint i=count; i; i -= real.sizeof)
                    {
                    * cast(real*) src = next.toReal();
                    src += real.sizeof;
                    }
                return count;
        }

        /***********************************************************************
        
        ***********************************************************************/

        override IReader get (inout char[] x, uint elements = uint.max) 
        { 
                // dup the input string, to avoid surprises
                x = next.toString (false);
                return this;
        }

        /***********************************************************************
        
                @todo - Tokenizer needs to handle wchar[] before this will
                        operate correctly

        ***********************************************************************/

        override IReader get (inout wchar[] x, uint elements = uint.max) 
        { 
                assert (0);
                return this;
        }

        /***********************************************************************
        
                @todo - Tokenizer needs to handle dchar[] before this will
                        operate correctly

        ***********************************************************************/

        override IReader get (inout dchar[] x, uint elements = uint.max) 
        { 
                assert (0);
                return this;
        }
}
